#!/usr/bin/env python

from online_cp.regressors import ConformalRidgeRegressor
from online_cp.classifiers import ConformalNearestNeighboursClassifier
from online_cp.martingale import PluginMartingale
from online_cp.CPS import RidgePredictionMachine


import numpy as np
import time
import warnings
from scipy.spatial.distance import pdist, cdist, squareform
from joblib import Parallel, delayed

class ConformalPredictionSet:

    def __init__(self, Gamma:np.array, epsilon):
        self.elements = Gamma
        self.epsilon = epsilon

    def __contains__(self, y):
        return y in self.elements
    
    def __len__(self):
        return self.elements.shape[0]
    
    def __repr__(self):
        return repr(self.elements)

    def __str__(self):
        return str(self.elements)

class ConformalClassifier:
    '''
    Parent class for classifiers
    '''

    def __init__(self):
        self.Err = 0
        # Preferred efficiency criteria (See Protocol 3.1 ALRW)
        self.OE = 0
        self.OF = 0


    @staticmethod
    def _compute_p_value(Alpha, tau=1, score_type='nonconformity', return_string=False):
        '''
        Assumes that the (non) conformity scores are organised so that the 
        test example is the last element.
        If tau is not provided, the non-smoothed p-value is computed.
        '''
        alpha_n = Alpha[-1]
        if score_type == 'nonconformity':
            gt = np.sum(Alpha > alpha_n)
            eq = np.sum(Alpha == alpha_n)
            p = (gt + tau * eq) / Alpha.shape[0]
            string = f'({gt} + {eq}*tau)/{Alpha.shape[0]}'

        elif score_type == 'conformity':
            lt = np.sum(Alpha < alpha_n)
            eq = np.sum(Alpha == alpha_n)
            p = (lt + tau * eq) / Alpha.shape[0]
            string = f'({lt} + {eq}*tau)/{Alpha.shape[0]}'
        
        if return_string:
            return float(p), string
        else:
            return float(p)


    def _compute_Gamma(self, p_values, epsilon):
        Gamma = []
        for y in self.label_space:
            if p_values[y] > epsilon:
                Gamma.append(y)
        return ConformalPredictionSet(np.array(Gamma), epsilon)
    

    def err(self, Gamma, y):
        err = int(not(y in Gamma))
        self.Err += err
        return err
    

    def oe(self, Gamma, y):
        if y in Gamma:
            oe = len(Gamma) - 1
        else:
            oe = len(Gamma)
        self.OE += oe
        return oe
    

    def of(self, p_values, y):
        of = 0
        for label, p in p_values.items():
            if not label == y:
                of += p
        self.OF += of
        return of

    # TEST
    def process_dataset(self, X, y, epsilon=0.1, init_train=0, return_results=False):

        self.label_space = np.unique(y)

        X_train = X[:init_train]
        y_train = y[:init_train]
        X_run = X[init_train:]
        y_run = y[init_train:]

        if return_results:
            res = np.zeros(shape=(y_run.shape[0], 3))
            prediction_sets = {}

        self.learn_initial_training_set(X=X_train, y=y_train)

        time_init = time.time()
        for i, (obj, lab) in enumerate(zip(X_run, y_run)):
            
            # Make prediction
            Gamma, p_values= self.predict(obj, epsilon=epsilon, return_p_values=True) 

            # Check error
            self.err(Gamma, lab)

            # Learn the label
            self.learn_one(obj, lab)
            
            # Prefferred efficiency criteria

            # Observed excess
            self.oe(Gamma, lab)

            # Observed fuzziness
            self.of(p_values, lab)

            if return_results:
                res[i, 0] = self.OE
                res[i, 1] = self.OF
                res[i, 2] = self.Err
                prediction_sets[i] = Gamma

        time_process = time.time() - time_init

        result = {
            'Efficiency': {
                'Average error': self.Err/self.y.shape[0],
                'Average OE': self.OE/self.y.shape[0],
                'Average OF': self.OF/self.y.shape[0],
                'Time': time_process
                }
            }
        if return_results:
            result['Prediction sets'] = prediction_sets,
            result['Cummulative Err'] = res[:, 2]
            result['Cummulative OE'] = res[:, 0]
            result['Cummulative OF'] = res[:, 1]
        
        return result
    
    

class ConformalNearestNeighboursClassifier(ConformalClassifier):
    """
    Classifier using nearest neighbours as the nonconformity measure.

    >>> cp = ConformalNearestNeighboursClassifier(k=1, rnd_state=1337)
    >>> Gamma, p_values = cp.predict(3, epsilon=0.1, return_p_values=True)
    >>> Gamma # predict both labels, as this is the first
    array([-1,  1])
    >>> [p_values[i] for i in [-1, 1]]
    [0.8781019003471183, 0.8781019003471183]

    >>> cp.learn_one(np.int64(3), 1)

    >>> Gamma, p_values = cp.predict(-2, epsilon=0.1, return_p_values=True)
    >>> Gamma # predict both labels, as this is the first
    array([-1,  1])
    >>> [p_values[i] for i in [-1, 1]]
    [0.18552796163759344, 0.18552796163759344]
    """
    # TODO: implement: cp.learn_several([[3,1],[4,7],[5,2]], [1, -1, 1])

    # TODO Write tests

    def __init__(self, k=1, label_space=np.array([-1, 1]), distance='euclidean', distance_func=None, verbose=0, rnd_state=None, n_jobs=None):
        super().__init__()
        self.label_space = label_space

        self.k = k

        self.distance = distance
        if distance_func is None:
            self.distance_func = self._standard_distance_func
        else:
            self.distance_func = distance_func
            self.distance = 'custom'

        self.y = np.empty(0)
        self.X = None
        self.D = None

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)

        self.n_jobs = n_jobs
    
    def reset(self):

        self.__init__(self.k, self.label_space)

    def _standard_distance_func(self, X, y=None):
        '''
        By default we use scipy to compute distances
        '''
        X = np.atleast_2d(X)
        if y is None:
            dists = squareform(pdist(X, metric=self.distance))
        else:
            y = np.atleast_2d(y)
            dists = cdist(X, y, metric=self.distance)
        return dists
    

    def learn_initial_training_set(self, X, y):
        if X.shape[0] > 0:
            self.X = X
            self.y = y
            self.D = self.distance_func(X)


    @staticmethod
    def update_distance_matrix(D, d):
        return np.block([[D, d], [d.T, np.array([0])]])
    

    def _find_nearest_distances(self, D, y):
        n = D.shape[0]
        
        # Initialize arrays to store the results
        same_label_distances = np.full(n, np.inf)
        different_label_distances = np.full(n, np.inf)

        for i in range(n):
            # Create a mask for the same and different labels
            same_label_mask = (y == y[i])
            different_label_mask = (y != y[i])

            # Ignore the distance to itself by setting it to np.inf
            same_label_mask[i] = False

            # Extract distances for the same label
            if np.any(same_label_mask):
                same_label_distances[i] = np.sort(D[i, same_label_mask])[:self.k].mean()
            
            # Extract distances for the different label
            if np.any(different_label_mask):
                different_label_distances[i] = np.sort(D[i, different_label_mask])[:self.k].mean()

        return same_label_distances, different_label_distances
    

    def learn_one(self, x, y, D=None):
        # Learn label y
        self.y = np.append(self.y, y)

        # Learn object
        if self.X is None:
            self.X = x.reshape(1,-1)
            self.D = self.distance_func(self.X)
        else:
            if D is None:
                d = self.distance_func(self.X, x)
                D = self.update_distance_matrix(self.D, d)
            self.D = D
            self.X = np.append(self.X, x.reshape(1, -1), axis=0)


    def predict(self, x, epsilon=0.1, return_p_values=False, return_update=False, verbose=0):
        p_values = {}
        tau = self.rnd_gen.uniform(0, 1)

        if self.y.shape[0] >= 1: 
            tic = time.time()
            d = self.distance_func(self.X, x)
            D = self.update_distance_matrix(self.D, d)
            time_update_D = time.time() - tic
            
            tic = time.time()
            if self.n_jobs is not None:
                def process_label(label):
                    y = np.append(self.y, label)
                    same_label_distances, different_label_distances = self._find_nearest_distances(D, y)

                    Alpha = same_label_distances / different_label_distances
                    if verbose > 10:
                        print(f'Nonconformity scores for hypothesis y={label}: {Alpha}')
                        _, string = self._compute_p_value(Alpha, tau, 'nonconformity', return_string=True)
                        print(f'p-value for hypothesis y={label}: {string}')

                    return label, self._compute_p_value(Alpha, tau, 'nonconformity')

                results = Parallel(n_jobs=self.n_jobs)(delayed(process_label)(label) for label in self.label_space)
                p_values = dict(results)
            else:
                for label in self.label_space:
                    y = np.append(self.y, label)
                    
                    same_label_distances, different_label_distances = self._find_nearest_distances(D, y)              

                    Alpha = np.nan_to_num(same_label_distances / different_label_distances, nan=np.inf)

                    if verbose > 10:
                        print(f'Nonconformity scores for hypothesis y={label}: {Alpha}')
                        p_values[label], string = self._compute_p_value(Alpha, tau, 'nonconformity', return_string=True)
                        print(f'p-value for hypothesis y={label}: {string}')

                    p_values[label] = self._compute_p_value(Alpha, tau, 'nonconformity')
            time_compute_p_values = time.time() - tic

            tic = time.time()
            Gamma = self._compute_Gamma(p_values, epsilon)
            time_Gamma = time.time()- tic

            self.time_dict = {
                'Update distance matrix': time_update_D,
                'Compute p-values': time_compute_p_values,
                'Compute Gamma': time_Gamma
            }
            
        else:
            for label in self.label_space:
                Alpha = np.array([np.inf])
                if verbose > 10:
                    print(f'Nonconformity scores for hypothesis y={label}: {Alpha}')
                    p_values[label], string = self._compute_p_value(Alpha, tau, 'nonconformity', return_string=True)
                    print(f'p-value for hypothesis y={label}: {string}')
                p_values[label] = self._compute_p_value(Alpha, tau, 'nonconformity')
            Gamma = self._compute_Gamma(p_values, epsilon)
            D = None
            self.time_dict = {}

        if return_update: 
            if return_p_values:
                return Gamma, p_values, D
            else:
                return Gamma, D
        else:
            if return_p_values:
                return Gamma, p_values
            else:
                return Gamma
        
    
    
if __name__ == "__main__":
    import doctest
    import sys
    (failures, _) = doctest.testmod()
    if failures:
        sys.exit(1)



# This file should contain the code to keep track of evaluation metrics
import numpy as np
import matplotlib.pyplot as plt

# NOTE: Below are some efficiency criteria, some of which are proper scoring rules.
#       What I have to do now, is to create one or more Evaluation classes, which, in their
#       construction create metric classes, and keeps track of the statistics as we go on.

import numpy as np

# This is a bit messy...
class Evaluation:

    def __init__(self, metrics=None, **kwargs):
        self.metrics = {}
        self.results = {}

        self.n = 0

        available_metrics = {
            "oe": OE,
            "of": OF,
            "err": Err,
            "width": Width,
            "winkler": WinklerScore,
            "crps": CRPS,
        }

        if metrics is not None:
            for metric_name in metrics:
                if metric_name.lower() in available_metrics:
                    self._add_metric_instance(metric_name.lower(), available_metrics[metric_name.lower()])
                else:
                    raise ValueError(f"Unknown metric: {metric_name}")

        for metric_name, metric_class in kwargs.items():
            if isinstance(metric_class, type) and issubclass(metric_class, (OE, OF, Err, Width, WinklerScore, CRPS)):
                self._add_metric_instance(metric_name.lower(), metric_class)
            elif metric_name.lower() in available_metrics:
                self._add_metric_instance(metric_name.lower(), available_metrics[metric_name.lower()])
            else:
                raise ValueError(f"Unknown metric: {metric_name}")

    def _add_metric_instance(self, name, metric_class):
        self.metrics[name] = metric_class()
        self.results[name] = []

    def update(self, y, Gamma=None, p_values=None, cpd=None, epsilon=None, raise_errors=True):
        for name, metric in self.metrics.items():
            try:
                if isinstance(metric, OE) or isinstance(metric, Err) or isinstance(metric, Width) or isinstance(metric, WinklerScore):
                    if Gamma is None:
                        if raise_errors:
                            raise ValueError(f"Gamma (gamma_val) is required for {name}.")
                        else:
                            continue

                    if isinstance(metric, WinklerScore):
                        result = metric._update(Gamma, y, epsilon)  # Correct arguments for WinklerScore
                    elif isinstance(metric, Width):
                        result = metric._update(Gamma) # Correct arguments for Width
                    else:
                        result = metric._update(y, Gamma)      # Correct arguments for OE, Err
                elif isinstance(metric, OF):
                    if p_values is None:
                        if raise_errors:
                            raise ValueError(f"Probability values (p_vals) are required for {name}.")
                        else:
                            continue
                    result = metric._update(p_values, y)           # Correct arguments for OF
                elif isinstance(metric, CRPS):
                    if cpd is None:
                        if raise_errors:
                            raise ValueError(f"CPD (cpd) is required for {name}.")
                        else:
                            continue
                    result = metric._update(cpd, y)              # Correct arguments for CRPS
                else:
                    raise ValueError(f"Unknown metric type: {type(metric)}")

                self.results[name].append(result)  # Store the result

            except ValueError as e:
                if raise_errors:
                    raise
                else:
                    print(f"Warning: Skipping metric '{name}': {e}")
        self.n += 1
        
    def summarize(self):
        summary = {}
        for name, results in self.results.items():
            summary[name] = {
                "mean": np.mean(results) if results else None,
                "std": np.std(results) if results else None,
                "min": np.min(results) if results else None,
                "max": np.max(results) if results else None,
                "last": results[-1] if results else None,
            }
        return summary
    
    def cumulative(self, criterion):
        return np.cumsum(self.results[criterion])
    
    def __getattr__(self, name):  # Called when an attribute is not found
        if name in self.metrics:
            return self.metrics[name]
        else:
            raise AttributeError(f"Metric '{name}' not found in results.")

    def __dir__(self): #For autocompletion
        return list(self.metrics.keys())
    
    
    # FIXME: I want to have x tics on the x-axis, but the same x-axis for all plots
    def plot_cumulative_results(self, max_cols=2, title=None):
        num_criteria = len(self.results)
        
        # Determine grid size
        num_cols = min(max_cols, num_criteria)  # Limit columns
        num_rows = int(np.ceil(num_criteria / num_cols))

        # Adaptive figure size
        fig, axs = plt.subplots(num_rows, num_cols, figsize=(5 * num_cols, 3 * num_rows), squeeze=False, sharex=True)

        # Flatten axs for easy indexing if it's multidimensional
        axs = axs.flatten()

        for i, (criterion, values) in enumerate(self.results.items()):
            cs = np.where(np.isinf(values), np.nan, np.cumsum(values))
            axs[i].plot(cs)
            axs[i].set_title(criterion)

        # Hide any unused subplots
        for j in range(i + 1, len(axs)):
            fig.delaxes(axs[j])

        if title is not None:
            fig.suptitle(title)
        fig.tight_layout()
        plt.close(fig)  # Prevent implicit display
        return fig


class EfficiencyCriterion:
    def __init__(self):
        self.value = 0.0
        self.n = 0

    def mean(self):
        return self.value / self.n

class OE(EfficiencyCriterion):

    def __init__(self):
        super().__init__()

    def _update(self, y, Gamma):
        self.n += 1
        if y in Gamma:
            oe = len(Gamma) - 1
        else:
            oe = len(Gamma)
        self.value += oe
        return oe
    
class OF(EfficiencyCriterion):

    def __init__(self):
        super().__init__()

    def _update(self, p_values, y):
        self.n += 1
        of = 0
        for label, p in p_values.items():
            if not label == y:
                of += p
        self.value += of
        return of
    
class Err(EfficiencyCriterion):

    def __init__(self):
        super().__init__()

    def _update(self, y, Gamma):
        self.n += 1
        if not type(Gamma) == tuple:
            err = int(not(y in Gamma))
        else:       
            err = int(not(y in Gamma))
        self.value += err
        return err

    
class Width(EfficiencyCriterion):

    def __init__(self):
        super().__init__()
    
    def _update(self, Gamma):
        self.n += 1
        w = Gamma.width()
        self.value += w
        return w
    
class WinklerScore(EfficiencyCriterion):

    def __init__(self):
        super().__init__()

    def _update(self, Gamma, y, epsilon=0.1):
        self.n += 1
        try:
            assert Gamma.upper < np.inf
            assert Gamma.lower > -np.inf
        except AssertionError:
            ws = np.inf
            self.value += ws
            return ws
        l = Gamma.lower
        u = Gamma.upper
        base_score = u - l
        penalty = 0
        
        if y < l:
            penalty = (2 / epsilon) * (l - y)
        elif y > u:
            penalty = (2 / epsilon) * (y - u)
        
        ws = base_score + penalty

        self.value += ws
        return ws
    
class CRPS(EfficiencyCriterion):

    def __init__(self):
        super().__init__()

    def _update(self, cpd, y):
        '''
        NOTE: This implementation is a bith shaky, really. I integrate the lower distribution for x <= y, 
              and the upper for x > y, to ensure convergence. It is possible that this is not proper.
        '''
        self.n += 1
        func_ad_hoc = lambda x: (cpd(x, 0) - int(y <= x))**2 if x <= y else (cpd(x, 1) - int(y <= x))**2

        crps = np.trapz([func_ad_hoc(x) for x in cpd.y_vals[1:-1]], cpd.y_vals[1:-1]) # NOTE: Perhaps not the best integrator

        self.value += crps
        return crps


import numpy as np
from scipy.integrate import quad
from scipy.optimize import minimize
from scipy.special import betaln
from scipy.stats import beta, gaussian_kde, norm
import warnings

class PluginMartingale:
    '''
    A conformal test martingale to test exchangeability online.
    We reject exchangeability with confidence 1-1/alpha if we ever observe M >= alpha for any alpha>0
    '''
    
    def __init__(self, method='kernel', window_size=None, warning_level=100, warnings=True, **kwargs):
        '''
        We can either use a kernel density estimate, or a parametric Beta model to estimate the density
        - method (str): The method for estimation ("kernel" of "beta")
        - kwargs: Additional parameters for the selected method.
            - For "kernel": "kernel_method" and "bandwidth"
            - For "beta": "mle" or "moments" or "bayes"

        TODO: Clear up the documentation. This could be messy.
        NOTE: It is possible to use a beta kernel for KDE. It is bounded by definition, and so, should have 
              no truoble with bounded data.
        TODO: Should we really have a parametric plugin? 
        '''
        self.logM = 0.0
        self.max = 1.0 # Keeps track of the maximum value so far.

        self.p_values = []
        self.log_martingale_values = [0.0] # NOTE: It may be better not to store the initial value...

        self.warning_level = warning_level
        self.warnings = warnings # Do we raise a user warning or not, when the warning level is reached

        self.method = method

        self.window_size = window_size

        self.params = kwargs

        # Initially, uniform betting function.
        self.b_n = lambda x: beta.pdf(x, 1, 1)
        self.B_n = lambda x: beta.cdf(x, 1, 1)

        self.min_sample_size = self.params.get("min_sample_size", 100) # After this many points, we use pure density estimate
        self.mixing_exponent = self.params.get("mixing_exponent", 1) # How fast does the mixing parameter grow to 1

        self.mixing_parameter = 0

        if self.method == 'kernel':
            self.kernel = self.params.get("kernel", 'gaussian' )
            if self.kernel == 'gaussian':
                self.kernel_method = self.params.get("kernel_method", 'reflect') # 'reflect' or 'logit'
                if self.kernel_method == 'logit':
                    self.edge_adjustment = self.params.get("edge_adjustment", 0.0)
            self.bandwidth = self.params.get("bandwidth", 'silverman') # 'silverman' or 'scott'

        elif self.method == 'beta':
            # Default is not to bet
            self.ahat = 1
            self.bhat = 1
            self.beta_method = self.params.get("beta_method", 'moment')

            if self.beta_method == 'moment':
                # Initialise online statistics
                self.n = 0  # Number of observations
                self.mean = 0.0  # Running mean
                self.M2 = 0.0  # Sum of squared differences from the mean (for variance)

            elif self.beta_method == 'mle':
                # Initialise online statistics
                self.n = 0  # Number of observations
                self.log_sum_x = 0.0
                self.log_sum_1_minus_x = 0.0

            elif self.beta_method == 'bayes':
                raise NotImplementedError('Working on this...')
                # FIXME: Figure out how to implement the Bayesian method. It is intuitively great to have the uniform distribution as the prior, 
                #        but I am confused as to how to implement it.
                self.prior_alpha = self.params.get("prior_alpha", 1)
                self.prior_beta = self.params.get("prior_beta", 1)
            else:
                raise Exception('beta_method must be one of "moment", "mle", bayes"')

    def update_mixing_parameter(self):
        self.mixing_parameter = min((len(self.p_values) / self.min_sample_size)**self.mixing_exponent, 1)

    @property
    def M(self):
        return np.exp(self.logM)
    
    def calculate_bandwidth(self, data, bw_mehtod='silverman', sigma=None):
        if bw_mehtod == 'silverman':
            assert sigma is not None
            h = ((4 * sigma**5) / (3 * data.size))**(1/5)
        else:
            raise NotImplementedError
        return h
    
    @property
    def martingale_values(self):
        return np.exp(self.log_martingale_values)
    
    @property
    def log10_martingale_values(self):
        return np.log10(self.martingale_values)
    
    # FIXME Need some calibration...
    def calculate_window_size(self):
        if self.window_size == 'adaptive':
            # FIXME DEFINITELY DO NOT WANT HARD CODED MIN SAMPLE SIZE
            window_param = np.log(1.001)
            min_size = self.min_sample_size
            max_size = len(self.p_values)
            return max(min_size, int(max_size*np.exp(-window_param*self.M)))
        elif self.window_size is None:
            return 0
        else:
            return self.window_size

    def update_martingale_value(self, p):
        if self.method == 'beta':
            self.logM += np.log(self.beta_betting_function(p))
        elif self.method == 'kernel':
            self.logM += np.log(self.kernel_betting_function(p))

        self.log_martingale_values.append(self.logM)
        self.p_values.append(p)

        self.update_mixing_parameter()

        if self.M > self.max:
            self.max = self.M

        if self.max >= self.warning_level and self.warnings:
            # TODO: Figure out how to warn only once!
            warnings.warn(f'Exchangeability assumption likely violated: Max martingale value is {self.max}')


    # FIXME: b_n must be set to 0 outside the unit interval, and B_n must be 0 to the left, and 1 to the right...
    def kernel_betting_function(self, p):
        if self.kernel == 'gaussian':
            gain, b_n, B_n = self.kernel_gaussian_betting_function(p)
            self.b_n = lambda x: self.mixing_parameter * b_n(x) + (1-self.mixing_parameter)
            self.B_n = lambda x: self.mixing_parameter * B_n(x) + (1-self.mixing_parameter)*x
            return self.mixing_parameter * gain + (1-self.mixing_parameter)
        else:
            raise NotImplementedError('There is currently only support for gaussian kernel')

    def kernel_gaussian_betting_function(self, p):

        if len(self.p_values) < 2:
            b_n = lambda x: beta.pdf(x, 1, 1)
            B_n = lambda x: beta.cdf(x, 1, 1)
            return 1, b_n, B_n
        else:
            data = np.array(self.p_values)[-self.calculate_window_size():]
            
            if self.kernel_method == 'reflect':
                # kde = gaussian_kde(data, bw_method=self.bandwidth)
                # h = kde.factor # The bandwidth

                sigma = data.std()

                # FIXME: Is this reasonable? Maybe rather something very concentrated
                if sigma == 0:
                    'If there is no variability: do not bet at all.'
                    b_n = lambda x: beta.pdf(x, 1, 1)
                    B_n = lambda x: beta.cdf(x, 1, 1)
                    return 1, b_n, B_n

                h = self.calculate_bandwidth(data=data, bw_mehtod=self.bandwidth, sigma=sigma)

                # Ensure `x` works for both scalars and arrays
                def kernel_pdf_raw(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    pdf_values = np.mean(norm.pdf(x[:, None], loc=data, scale=h * sigma), axis=1)
                    if pdf_values.size == 1:  # Check if the result is a single value
                        return pdf_values.item()  # Convert single-element array to scalar
                    return pdf_values

                def kernel_cdf_raw(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    cdf_values = np.mean(norm.cdf(x[:, None], loc=data, scale=h * sigma), axis=1)
                    if cdf_values.size == 1:  # Check if the result is a single value
                        return cdf_values.item()  # Convert single-element array to scalar
                    return cdf_values

                def kernel_pdf_reflect(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    pdf_reflect_values = kernel_pdf_raw(x) + kernel_pdf_raw(-x) + kernel_pdf_raw(2 - x)
                    if np.isscalar(pdf_reflect_values):  # Check if the result is scalar
                        return pdf_reflect_values
                    return pdf_reflect_values

                def kernel_cdf_reflect(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    cdf_reflect_values = kernel_cdf_raw(x) - kernel_cdf_raw(-x) + 1 - kernel_cdf_raw(2 - x)
                    if np.isscalar(cdf_reflect_values):  # Check if the result is scalar
                        return cdf_reflect_values
                    return cdf_reflect_values

                return kernel_pdf_reflect(p), kernel_pdf_reflect, kernel_cdf_reflect
            
            elif self.kernel_method == 'logit':
                def logit(x):
                    return np.log(x / (1-x))
                def edge_adjustment(x, delta=0.001):
                    return delta + (1-2*delta)*x
                def transform_with_edge_adjustment(x, delta=0.001):
                    return logit(edge_adjustment(x, delta))
                def derivative_transform_with_edge_adjustment(x, delta=0.001):
                    return (2*delta - 1) / ((2*delta*x - delta - x)*(2*delta*x - delta - x + 1))
                
                transformed_data = transform_with_edge_adjustment(data, self.edge_adjustment)

                # kde = gaussian_kde(transformed_data, bw_method=self.bandwidth)
                # h = kde.factor
                sigma = transformed_data.std()

                if sigma == 0:
                    'If there is no variability: do not bet at all.'
                    b_n = lambda x: beta.pdf(x, 1, 1)
                    B_n = lambda x: beta.cdf(x, 1, 1)
                    return 1, b_n, B_n


                h = self.calculate_bandwidth(data=data, bw_mehtod=self.bandwidth, sigma=sigma)

                def kernel_pdf_transformed(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    pdf_values = np.mean(norm.pdf(x[:, None], loc=transformed_data, scale=h * sigma), axis=1)
                    if pdf_values.size == 1:  # Check if the result is a single value
                        return pdf_values.item()  # Convert single-element array to scalar
                    return pdf_values

                def kernel_cdf_transformed(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    cdf_values = np.mean(norm.cdf(x[:, None], loc=transformed_data, scale=h * sigma), axis=1)
                    if cdf_values.size == 1:  # Check if the result is a single value
                        return cdf_values.item()  # Convert single-element array to scalar
                    return cdf_values

                def kernel_pdf_logit(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    pdf_transform_values = kernel_pdf_transformed(transform_with_edge_adjustment(x, self.edge_adjustment)) * derivative_transform_with_edge_adjustment(x, self.edge_adjustment)
                    if pdf_transform_values.shape[0] == 1: 
                        return pdf_transform_values[0]
                    return pdf_transform_values

                def kernel_cdf_logit(x):
                    x = np.atleast_1d(x)  # Ensure x is an array
                    cdf_transform_values = kernel_cdf_transformed(transform_with_edge_adjustment(x, self.edge_adjustment))
                    return cdf_transform_values
                
                return kernel_pdf_logit(p), kernel_pdf_logit, kernel_cdf_logit
            
            else:
                raise NotImplementedError(f'Kernel method {self.kernel_method} does not exist.')


    def beta_betting_function(self, p):
        if self.beta_method == 'moment':
            gain = self.beta_moment_betting_function(p) # Is gain really a good name? I mean the outcome of the bet...
            self.b_n = lambda x: self.mixing_parameter * beta.pdf(x, self.ahat, self.bhat) + (1-self.mixing_parameter )
            self.B_n = lambda x: self.mixing_parameter * beta.cdf(x, self.ahat, self.bhat) + (1-self.mixing_parameter )*x
            return self.mixing_parameter * gain + (1-self.mixing_parameter)
            
        elif self.beta_method == 'mle':
            gain = self.beta_mle_betting_function(p) # Is gain really a good name? I mean the outcome of the bet...
            self.b_n = lambda x: self.mixing_parameter * beta.pdf(x, self.ahat, self.bhat) + (1-self.mixing_parameter )
            self.B_n = lambda x: self.mixing_parameter * beta.cdf(x, self.ahat, self.bhat) + (1-self.mixing_parameter )*x
            return self.mixing_parameter * gain + (1-self.mixing_parameter)
        
        elif self.beta_method == 'bayes':
            raise NotImplementedError()
        
    def beta_moment_betting_function(self, p):
        if self.n < 2:
            self.ahat = 1.0  # Parameters undefined, choose uniform as default
            self.bhat = 1.0
        
        else:
            sample_variance = self.M2 / (self.n - 1) if self.n > 1 else 0
            
            if sample_variance <= 0:
                self.ahat = 1  # Parameters undefined, choose uniform as default
                self.bhat = 1
            
            else:
                common_factor = (self.mean * (1 - self.mean) / sample_variance) - 1
                self.ahat = self.mean * common_factor
                self.bhat = (1 - self.mean) * common_factor

        # Update statistics
        self.n += 1
        delta = p - self.mean
        self.mean += delta / self.n
        delta2 = p - self.mean
        self.M2 += delta * delta2  # Incremental update for variance

        return beta.pdf(p, self.ahat, self.bhat)
    
    def beta_mle_betting_function(self, p):
        def negative_log_likelihood(params):
            alpha, beta = params
            if alpha <= 0 or beta <= 0:
                return np.inf  # Invalid parameters
            log_likelihood = (
                (alpha - 1) * self.log_sum_x +
                (beta - 1) * self.log_sum_1_minus_x -
                self.n * betaln(alpha, beta)
            )
            return -log_likelihood  # Negate for minimization

        if self.n < 2:
            self.ahat = 1.0  # Parameters undefined, choose uniform as default
            self.bhat = 1.0
        else:
            initial_guess = [1.0, 1.0]
            result = minimize(negative_log_likelihood, initial_guess,
                            bounds=[(1e-5, None), (1e-5, None)])
            if result.success:
                self.ahat, self.bhat =  tuple(result.x)  # Optimal alpha, beta
            else:
                self.ahat, self.bhat = 1.0, 1.0 # If we can not say anything, choose uniform as default
            
        # Update statistics
        self.n += 1
        self.log_sum_x += np.log(p)
        self.log_sum_1_minus_x += np.log(1 - p)

        return beta.pdf(p, self.ahat, self.bhat)

if __name__ == "__main__":
    import doctest
    import sys
    (failures, _) = doctest.testmod()
    if failures:
        sys.exit(1)
